<?php

declare(strict_types=1);

namespace Drupal\user_registration_reminder;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\token\Token;
use Drupal\user\UserDataInterface;

/**
 * User Registration Reminder Manager class, handling the business logic.
 */
class Manager {
  const FINAL_ACTION_NONE = 'none';
  const FINAL_ACTION_BLOCK = 'block';
  const FINAL_ACTION_DELETE = 'delete';

  use StringTranslationTrait;

  /**
   * The module settings.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected readonly ImmutableConfig $config;

  /**
   * The module logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected readonly LoggerChannelInterface $logger;

  /**
   * Constructs a Manager object.
   */
  public function __construct(
    protected readonly ConfigFactoryInterface $configFactory,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly MailManagerInterface $mailManager,
    protected readonly UserDataInterface $userData,
    protected readonly LoggerChannelFactoryInterface $loggerChannelFactory,
    protected readonly Token $token,
    protected readonly TimeInterface $timeService,
  ) {
    $this->config = $configFactory->get('user_registration_reminder.settings');
    $this->logger = $loggerChannelFactory->get('user_registration_reminder');
  }

  /**
   * Runs the registration reminder check and processing.
   */
  public function runRegistrationReminder(): void {
    $inactiveRegistrationUsers = $this->getInactiveRegistrationUsers();
    foreach ($inactiveRegistrationUsers as $inactiveRegistrationUser) {
      /** @var UserDecorator $inactiveRegistrationUser */
      $nextEscalationStep = $inactiveRegistrationUser->getLastEscalationStep() + 1;
      $lastEscalationTimestamp = $inactiveRegistrationUser->getLastEscalationTimestamp();
      if ($nextEscalationStep >= $this->getFinalEscalationStepId()) {
        // Final escalation step:
        if ($this->getFinalActionDelay() > 0 && $lastEscalationTimestamp + $this->getFinalActionDelay() <= time()) {
          $inactiveRegistrationUser->executeFinalAction($this->getFinalActionAction());
        }
      }
      else {
        // Regular escalation step before the final escalation:
        $escalationStepConfig = $this->getEscalationStepConfig($nextEscalationStep);
        if (!$escalationStepConfig) {
          // Escalation step configuration expected, but none found.
          // Log an error and skip:
          $this->logger->debug($this->t('No escalation step configuration found for escalation step: %step. No registration reminder sent for user with UID: %uid', [
            '%step' => $nextEscalationStep,
            '%uid' => $inactiveRegistrationUser->getUser()->id(),
          ]));
          continue;
        }
        if ($lastEscalationTimestamp + $escalationStepConfig['delay'] <= time()) {
          if ($escalationStepConfig['delay'] !== 0) {
            // Only send mail if not disabled by delay = 0.
            $this->sendReminderEmail($inactiveRegistrationUser);
          }
          $inactiveRegistrationUser->increaseEscalationStep();
        }
      }

    }
  }

  /**
   * Returns the final escalation step ID BEFORE the final action.
   *
   * E.g. if 3 reminder steps are defined, returns "3".
   * Only counts steps with duration > 0.
   *
   * @return int
   *   The active escalation step count.
   */
  protected function getFinalEscalationStepCount(): int {
    $i = 0;
    $reminderEscalations = $this->config->get("reminder_escalations");
    foreach ($reminderEscalations as $reminderEscalation) {
      if ($reminderEscalation['duration'] > 0) {
        $i++;
      }
    }
    return $i;
  }

  /**
   * Returns the config array from the given escalation step.
   *
   * Returns NULL, if the requested escalation step ist not existing.
   *
   * @param int $escalationStep
   *   The escalation step count.
   *
   * @return array|null
   *   The escalation step configuration or NULL if not existing.
   */
  protected function getEscalationStepConfig(int $escalationStep): ?array {
    return $this->config->get("reminder_escalations.{$escalationStep}") ?? NULL;
  }

  /**
   * Returns the final action delay in seconds.
   *
   * The final action is run with the given delay, after the last escalation
   * step.
   *
   * @return int
   *   The final action delay after the last escalation step.
   */
  protected function getFinalActionDelay(): int {
    return $this->config->get('final_action.delay') ?? 0;
  }

  /**
   * Returns the final action to take after the last escalation step.
   *
   * The final action is run with the given delay, after the last escalation
   * step.
   *
   * @return string
   *   The final action delay after the last escalation step.
   */
  protected function getFinalActionAction(): string {
    return $this->config->get('final_action.action') ?? 'none';
  }

  /**
   * Helper method to return the timestamp of the very first escalation.
   *
   * All users that were created *before* this timestamp and match the other
   * conditions, should be handled in some way.
   *
   * @return int
   *   The first escalation timestamp to compare user creation against.
   */
  protected function getFirstEscalationTimestamp(): int {
    $delay = $this->config->get('reminder_escalations.0.delay');
    $time = $this->timeService->getRequestTime();
    return $time - $delay;
  }

  /**
   * Returns the users that never logged in.
   *
   * Uses UserDecorator to provide further reminder information.
   *
   * @return \Drupal\user\Entity\UserDecorator[]
   *   The array of UserDecorator objects.
   */
  protected function getInactiveRegistrationUsers(): array {
    $createdBeforeTimestamp = $this->getFirstEscalationTimestamp();
    $query = $this->entityTypeManager->getStorage('user')->getQuery();
    $uids = $query
    // Only enabled users:
      ->condition('status', 1)
      // Only users who never logged in.
      ->condition('login', 0)
      ->condition('created', $createdBeforeTimestamp, '<')
      ->execute();

    if (empty($uids)) {
      return [];
    }
    $users = [];
    foreach ($uids as $uid) {
      $user = $this->entityTypeManager->getStorage('user')->load($uid);
      $users[] = new UserDecorator($user, $this->userData);
    }
    return $users;
  }

  /**
   * Helper method to send the reminder email.
   *
   * @param UserDecorator $inactiveRegistrationUser
   *   The user to send the reminder to.
   *
   * @return bool
   *   Returns TRUE if the mail was sent successful, FALSE otherwise.
   */
  protected function sendReminderEmail(UserDecorator $inactiveRegistrationUser): bool {
    $user = $inactiveRegistrationUser->getUser();
    $email = $user->getEmail();
    $escalationStep = $inactiveRegistrationUser->getLastEscalationStep() + 1;
    $escalationStepConfig = $this->getEscalationStepConfig($escalationStep);
    $subject = $escalationStepConfig['email_subject'];
    $body = $escalationStepConfig['email_body'];
    $tokenValues = [
      'user' => $user,
    ];
    $subject = $this->token->replace($subject, $tokenValues, ['clear' => TRUE]);
    $subject = Html::decodeEntities(strip_tags($subject));
    $body = $this->token->replace($body, $tokenValues, ['clear' => TRUE]);
    $body = Html::decodeEntities(strip_tags($body));
    $params = [
      'subject' => $subject,
      'body' => $body,
    ];
    return $this->mailManager->mail(
      'user_registration_reminder',
      $inactiveRegistrationUser->getUser()->id(),
      $email,
      $user->getPreferredLangcode(),
      $params,
    )['result'] ?? TRUE;
  }

}
